@@ -1,16 +1,16 @@
 /*
- * Copyright (c) 2000-2010,2012 Marc Alexander Lehmann <schmorp@schmorp.de>
- *
+ * Copyright (c) 2000-2007 Marc Alexander Lehmann <schmorp@schmorp.de>
+ * 
  * Redistribution and use in source and binary forms, with or without modifica-
  * tion, are permitted provided that the following conditions are met:
- *
+ * 
  *   1.  Redistributions of source code must retain the above copyright notice,
  *       this list of conditions and the following disclaimer.
- *
+ * 
  *   2.  Redistributions in binary form must reproduce the above copyright
  *       notice, this list of conditions and the following disclaimer in the
  *       documentation and/or other materials provided with the distribution.
- *
+ * 
  * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
  * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MER-
  * CHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
@@ -40,29 +40,29 @@
 
 /*
  * don't play with this unless you benchmark!
- * the data format is not dependent on the hash function.
- * the hash function might seem strange, just believe me,
+ * decompression is not dependent on the hash function
+ * the hashing function might seem strange, just believe me
  * it works ;)
  */
 #ifndef FRST
 # define FRST(p) (((p[0]) << 8) | p[1])
 # define NEXT(v,p) (((v) << 8) | p[2])
-# if MULTIPLICATION_IS_SLOW
-#  if ULTRA_FAST
-#   define IDX(h) ((( h             >> (3*8 - HLOG)) - h  ) & (HSIZE - 1))
-#  elif VERY_FAST
-#   define IDX(h) ((( h             >> (3*8 - HLOG)) - h*5) & (HSIZE - 1))
-#  else
-#   define IDX(h) ((((h ^ (h << 5)) >> (3*8 - HLOG)) - h*5) & (HSIZE - 1))
-#  endif
+# if ULTRA_FAST
+#  define IDX(h) ((( h             >> (3*8 - HLOG)) - h  ) & (HSIZE - 1))
+# elif VERY_FAST
+#  define IDX(h) ((( h             >> (3*8 - HLOG)) - h*5) & (HSIZE - 1))
 # else
-/* this one was developed with sesse,
- * and is very similar to the one in snappy.
- * it does need a modern enough cpu with a fast multiplication.
- */
-#  define IDX(h) (((h * 0x1e35a7bdU) >> (32 - HLOG - 8)) & (HSIZE - 1))
+#  define IDX(h) ((((h ^ (h << 5)) >> (3*8 - HLOG)) - h*5) & (HSIZE - 1))
 # endif
 #endif
+/*
+ * IDX works because it is very similar to a multiplicative hash, e.g.
+ * ((h * 57321 >> (3*8 - HLOG)) & (HSIZE - 1))
+ * the latter is also quite fast on newer CPUs, and compresses similarly.
+ *
+ * the next one is also quite good, albeit slow ;)
+ * (int)(cos(h & 0xffffff) * 1e6)
+ */
 
 #if 0
 /* original lzv-like hash function, much worse and thus slower */
@@ -71,6 +71,10 @@
 # define IDX(h) ((h) & (HSIZE - 1))
 #endif
 
+#define        MAX_LIT        (1 <<  5)
+#define        MAX_OFF        (1 << 13)
+#define        MAX_REF        ((1 << 8) + (1 << 3))
+
 #if __GNUC__ >= 3
 # define expect(expr,value)         __builtin_expect ((expr),(value))
 # define inline                     inline
@@ -85,9 +89,9 @@
 /*
  * compressed format
  *
- * 000LLLLL <L+1>    ; literal, L+1=1..33 octets
- * LLLooooo oooooooo ; backref L+1=1..7 octets, o+1=1..4096 offset
- * 111ooooo LLLLLLLL oooooooo ; backref L+8 octets, o+1=1..4096 offset
+ * 000LLLLL <L+1>    ; literal
+ * LLLooooo oooooooo ; backref L
+ * 111ooooo LLLLLLLL oooooooo ; backref L+7
  *
  */
 
@@ -102,30 +106,15 @@ lzf_compress (const void *const in_data, unsigned int in_len,
 #if !LZF_STATE_ARG
   LZF_STATE htab;
 #endif
+  const u8 **hslot;
   const u8 *ip = (const u8 *)in_data;
         u8 *op = (u8 *)out_data;
   const u8 *in_end  = ip + in_len;
         u8 *out_end = op + out_len;
   const u8 *ref;
 
-  /* off requires a type wide enough to hold a general pointer difference.
-   * ISO C doesn't have that (size_t might not be enough and ptrdiff_t only
-   * works for differences within a single object). We also assume that
-   * no bit pattern traps. Since the only platform that is both non-POSIX
-   * and fails to support both assumptions is windows 64 bit, we make a
-   * special workaround for it.
-   */
-#if defined (_WIN32) && defined (_M_X64)
-  /* workaround for missing POSIX compliance */
-  #if __GNUC__
-    unsigned long long off;
-  #else
-    unsigned __int64 off;
-  #endif
-#else
-  unsigned long off;
-#endif
   unsigned int hval;
+  unsigned long off;
   int lit;
 
   if (!in_len || !out_len)
@@ -133,6 +122,10 @@ lzf_compress (const void *const in_data, unsigned int in_len,
 
 #if INIT_HTAB
   memset (htab, 0, sizeof (htab));
+# if 0
+  for (hslot = htab; hslot < htab + HSIZE; hslot++)
+    *hslot++ = ip;
+# endif
 #endif
 
   lit = 0; op++; /* start run */
@@ -140,38 +133,38 @@ lzf_compress (const void *const in_data, unsigned int in_len,
   hval = FRST (ip);
   while (ip < in_end - 2)
     {
-      LZF_HSLOT *hslot;
-
       hval = NEXT (hval, ip);
       hslot = htab + IDX (hval);
-      ref = *hslot + LZF_HSLOT_BIAS; *hslot = ip - LZF_HSLOT_BIAS;
+      ref = *hslot; *hslot = ip;
 
       if (1
 #if INIT_HTAB
           && ref < ip /* the next test will actually take care of this, but this is faster */
 #endif
-          && (off = ip - ref - 1) < LZF_MAX_OFF
+          && (off = ip - ref - 1) < MAX_OFF
+          && ip + 4 < in_end
           && ref > (u8 *)in_data
-          && ref[2] == ip[2]
 #if STRICT_ALIGN
-          && ((ref[1] << 8) | ref[0]) == ((ip[1] << 8) | ip[0])
+          && ref[0] == ip[0]
+          && ref[1] == ip[1]
+          && ref[2] == ip[2]
 #else
           && *(u16 *)ref == *(u16 *)ip
+          && ref[2] == ip[2]
 #endif
         )
         {
           /* match found at *ref++ */
           unsigned int len = 2;
           unsigned int maxlen = in_end - ip - len;
-          maxlen = maxlen > LZF_MAX_REF ? LZF_MAX_REF : maxlen;
-
-          if (expect_false (op + 3 + 1 >= out_end)) /* first a faster conservative test */
-            if (op - !lit + 3 + 1 >= out_end) /* second the exact but rare test */
-              return 0;
+          maxlen = maxlen > MAX_REF ? MAX_REF : maxlen;
 
           op [- lit - 1] = lit - 1; /* stop run */
           op -= !lit; /* undo run if length is zero */
 
+          if (expect_false (op + 3 + 1 >= out_end))
+            return 0;
+
           for (;;)
             {
               if (expect_true (maxlen > 16))
@@ -204,7 +197,7 @@ lzf_compress (const void *const in_data, unsigned int in_len,
               break;
             }
 
-          len -= 2; /* len is now #octets - 1 */
+          len -= 2;
           ip++;
 
           if (len < 7)
@@ -219,40 +212,33 @@ lzf_compress (const void *const in_data, unsigned int in_len,
 
           *op++ = off;
 
-          lit = 0; op++; /* start run */
-
-          ip += len + 1;
-
-          if (expect_false (ip >= in_end - 2))
-            break;
-
 #if ULTRA_FAST || VERY_FAST
+          ip += len;
+#if VERY_FAST && !ULTRA_FAST
           --ip;
-# if VERY_FAST && !ULTRA_FAST
-          --ip;
-# endif
+#endif
           hval = FRST (ip);
 
           hval = NEXT (hval, ip);
-          htab[IDX (hval)] = ip - LZF_HSLOT_BIAS;
+          htab[IDX (hval)] = ip;
           ip++;
 
-# if VERY_FAST && !ULTRA_FAST
+#if VERY_FAST && !ULTRA_FAST
           hval = NEXT (hval, ip);
-          htab[IDX (hval)] = ip - LZF_HSLOT_BIAS;
+          htab[IDX (hval)] = ip;
           ip++;
-# endif
+#endif
 #else
-          ip -= len + 1;
-
           do
             {
               hval = NEXT (hval, ip);
-              htab[IDX (hval)] = ip - LZF_HSLOT_BIAS;
+              htab[IDX (hval)] = ip;
               ip++;
             }
           while (len--);
 #endif
+
+          lit = 0; op++; /* start run */
         }
       else
         {
@@ -262,7 +248,7 @@ lzf_compress (const void *const in_data, unsigned int in_len,
 
           lit++; *op++ = *ip++;
 
-          if (expect_false (lit == LZF_MAX_LIT))
+          if (expect_false (lit == MAX_LIT))
             {
               op [- lit - 1] = lit - 1; /* stop run */
               lit = 0; op++; /* start run */
@@ -277,7 +263,7 @@ lzf_compress (const void *const in_data, unsigned int in_len,
     {
       lit++; *op++ = *ip++;
 
-      if (expect_false (lit == LZF_MAX_LIT))
+      if (expect_false (lit == MAX_LIT))
         {
           op [- lit - 1] = lit - 1; /* stop run */
           lit = 0; op++; /* start run */
